<?php

namespace Monarch;

/**
 * HSL color class.
 */
class HSL extends AccessibleColor {

  /**
   * The hue component.
   *
   * @var float
   */
  protected $hue = 0.0;

  /**
   * The saturation component.
   *
   * @var float
   */
  protected $saturation = 0.0;

  /**
   * The lightness component.
   *
   * @var float
   */
  protected $lightness = 0.0;

  /**
   * Factory function.
   *
   * @param float $hue
   *   The hue component.
   * @param float $saturation
   *   The saturation component.
   * @param float $lightness
   *   The lightness component.
   *
   * @return HSL
   *   The created HSL object.
   */
  public static function create(float $hue, float $saturation, float $lightness) : HSL {
    return new static($hue, $saturation, $lightness);
  }

  /**
   * {@inheritdoc}
   */
  public static function fromHex(string $hex) : HSL {
    return RGB::fromHex($hex)->toHsl();
  }

  /**
   * Constructor function.
   *
   * @param float $hue
   *   The hue component.
   * @param float $saturation
   *   The saturation component.
   * @param float $lightness
   *   The lightness component.
   */
  public function __construct(float $hue, float $saturation, float $lightness) {
    $this->hue = $hue;
    $this->saturation = $saturation;
    $this->lightness = $lightness;
  }

  /**
   * Wrap and return the hue value.
   *
   * @return int
   *   The clamped and casted value.
   */
  public function h() : float {
    return fmod(fmod($this->hue, 360.0) + 360.0, 360.0);
  }

  /**
   * Clamp and return the saturation value.
   *
   * @return int
   *   The clamped and casted value.
   */
  public function s() : float {
    return max(0.0, min(1.0, $this->saturation));
  }

  /**
   * Clamp and return the lightness value.
   *
   * @return int
   *   The clamped and casted value.
   */
  public function l() : float {
    return max(0.0, min(1.0, $this->lightness));
  }

  /**
   * {@inheritdoc}
   */
  public function toRgb() : RGB {
    $hue = $this->h();
    $saturation = $this->s();
    $lightness = $this->l();

    $c = (1.0 - abs(2.0 * $lightness - 1.0)) * $saturation;
    $x = $c * (1.0 - abs(fmod(($hue / 60.0), 2.0) - 1.0));
    $m = $lightness - $c / 2.0;

    [$dr, $dg, $db] = [
      [$c, $x, 0.0],
      [$x, $c, 0.0],
      [0.0, $c, $x],
      [0.0, $x, $c],
      [$x, 0.0, $c],
      [$c, 0.0, $x],
    ][(int) ($hue / 60.0)];

    return RGB::create(($dr + $m) * 255.0, ($dg + $m) * 255.0, ($db + $m) * 255.0);
  }

  /**
   * {@inheritdoc}
   */
  public function toHsl() : HSL {
    return new HSL($this->hue, $this->saturation, $this->lightness);
  }

  /**
   * Convert to string.
   *
   * @return string
   *   The string representation of the class.
   */
  public function __toString() : string {
    return 'hsl(' . implode(', ', [
      $this->h(),
      ($this->s() * 100) . '%',
      ($this->l() * 100) . '%',
    ]) . ')';
  }

  /**
   * {@inheritdoc}
   */
  public function toHex() : string {
    return $this->toRgb()->toHex();
  }

  /**
   * {@inheritdoc}
   */
  public function getLuminance() : float {
    return $this->toRgb()->getLuminance();
  }

  /**
   * {@inheritdoc}
   */
  public function getLightness() : float {
    return $this->l();
  }

  /**
   * {@inheritdoc}
   */
  public function setLightness(float $lightness) : HSL {
    $this->lightness = $lightness;

    return $this;
  }

}
